@@ -280,9 +280,9 @@ class Parser (Recognizer):
     def compileParseTreePattern(self, pattern, patternRuleIndex, lexer = None):
         if lexer is None:
             if self.getTokenStream() is not None:
-                tokenSource = self.getTokenStream().getTokenSource()
-            if isinstance( tokenSource, Lexer ):
-                lexer = tokenSource
+                tokenSource = self.getTokenStream().tokenSource
+                if isinstance( tokenSource, Lexer ):
+                    lexer = tokenSource
         if lexer is None:
             raise UnsupportedOperationException("Parser can't discover a lexer to use")
 
