@@ -73,6 +73,9 @@ class PredictionContext(object):
     def __init__(self, cachedHashCode):
         self.cachedHashCode = cachedHashCode
 
+    def __len__(self):
+        return 0
+
     # This means only the {@link #EMPTY} context is in set.
     def isEmpty(self):
         return self is self.EMPTY
@@ -80,6 +83,9 @@ class PredictionContext(object):
     def hasEmptyPath(self):
         return self.getReturnState(len(self) - 1) == self.EMPTY_RETURN_STATE
 
+    def getReturnState(self, index):
+        raise "illegal!"
+
     def __hash__(self):
         return self.cachedHashCode
 
@@ -88,11 +94,13 @@ class PredictionContext(object):
 
 
 def calculateHashCode(parent, returnState):
-    return hash( str(parent) + str(returnState))
-
-def calculateEmptyHashCode():
-    return hash("")
+    return hash("") if parent is None else hash((hash(parent), returnState))
 
+def calculateListsHashCode(parents, returnStates ):
+    h = 0
+    for parent, returnState in parents, returnStates:
+        h = hash((h, calculateHashCode(parent, returnState)))
+    return h
 
 #  Used to cache {@link PredictionContext} objects. Its used for the shared
 #  context cash associated with contexts in DFA states. This cache
@@ -134,8 +142,7 @@ class SingletonPredictionContext(PredictionContext):
             return SingletonPredictionContext(parent, returnState)
 
     def __init__(self, parent, returnState):
-        assert returnState!=ATNState.INVALID_STATE_NUMBER
-        hashCode = calculateHashCode(parent, returnState) if parent is not None else calculateEmptyHashCode()
+        hashCode = calculateHashCode(parent, returnState)
         super(SingletonPredictionContext, self).__init__(hashCode)
         self.parentCtx = parent
         self.returnState = returnState
@@ -144,11 +151,9 @@ class SingletonPredictionContext(PredictionContext):
         return 1
 
     def getParent(self, index):
-        assert index == 0
         return self.parentCtx
 
     def getReturnState(self, index):
-        assert index == 0
         return self.returnState
 
     def __eq__(self, other):
@@ -158,8 +163,6 @@ class SingletonPredictionContext(PredictionContext):
             return False
         elif not isinstance(other, SingletonPredictionContext):
             return False
-        elif hash(self) != hash(other):
-            return False # can't be same if hash is different
         else:
             return self.returnState == other.returnState and self.parentCtx==other.parentCtx
 
@@ -185,15 +188,12 @@ class EmptyPredictionContext(SingletonPredictionContext):
     def isEmpty(self):
         return True
 
-    def getParent(self, index):
-        return None
-
-    def getReturnState(self, index):
-        return self.returnState
-
     def __eq__(self, other):
         return self is other
 
+    def __hash__(self):
+        return self.cachedHashCode
+
     def __unicode__(self):
         return "$"
 
@@ -206,9 +206,7 @@ class ArrayPredictionContext(PredictionContext):
     #  returnState == {@link #EMPTY_RETURN_STATE}.
 
     def __init__(self, parents, returnStates):
-        super(ArrayPredictionContext, self).__init__(calculateHashCode(parents, returnStates))
-        assert parents is not None and len(parents)>0
-        assert returnStates is not None and len(returnStates)>0
+        super(ArrayPredictionContext, self).__init__(calculateListsHashCode(parents, returnStates))
         self.parents = parents
         self.returnStates = returnStates
 
@@ -276,17 +274,7 @@ def PredictionContextFromRuleContext(atn, outerContext=None):
     return SingletonPredictionContext.create(parent, transition.followState.stateNumber)
 
 
-def calculateListsHashCode(parents, returnStates ):
-
-    with StringIO() as s:
-        for parent in parents:
-            s.write(unicode(parent))
-        for returnState in returnStates:
-            s.write(unicode(returnState))
-        return hash(s.getvalue())
-
 def merge(a, b, rootIsWildcard, mergeCache):
-    assert a is not None and b is not None # must be empty context, never null
 
     # share same graph if both same
     if a==b:
@@ -305,9 +293,9 @@ def merge(a, b, rootIsWildcard, mergeCache):
 
     # convert singleton so both are arrays to normalize
     if isinstance( a, SingletonPredictionContext ):
-        a = ArrayPredictionContext(a)
+        a = ArrayPredictionContext([a.parentCtx], [a.returnState])
     if isinstance( b, SingletonPredictionContext):
-        b = ArrayPredictionContext(b)
+        b = ArrayPredictionContext([b.parentCtx], [b.returnState])
     return mergeArrays(a, b, rootIsWildcard, mergeCache)
 
 
@@ -380,7 +368,7 @@ def mergeSingletons(a, b, rootIsWildcard, mergeCache):
                 payloads[0] = b.returnState
                 payloads[1] = a.returnState
             parents = [singleParent, singleParent]
-            a_ = ArrayPredictionContext(parents, payloads);
+            a_ = ArrayPredictionContext(parents, payloads)
             if mergeCache is not None:
                 mergeCache.put(a, b, a_)
             return a_
@@ -486,9 +474,9 @@ def mergeArrays(a, b, rootIsWildcard, mergeCache):
             return previous
 
     # merge sorted payloads a + b => M
-    i = 0; # walks a
-    j = 0; # walks b
-    k = 0; # walks target M array
+    i = 0 # walks a
+    j = 0 # walks b
+    k = 0 # walks target M array
 
     mergedReturnStates = [] * (len(a.returnState) + len( b.returnStates))
     mergedParents = [] * len(mergedReturnStates)
